 package baldr;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.HashSet;
import java.util.Iterator;

import org.apache.commons.collections.buffer.CircularFifoBuffer;
import org.apache.commons.collections.buffer.UnboundedFifoBuffer;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.SimpleAnalyzer;
import org.apache.lucene.analysis.Token;
import org.apache.lucene.analysis.TokenFilter;
import org.apache.lucene.analysis.TokenStream;

/**
 * An incapsulating class which holds differnt kinds of TokenFilters:
 * <ul>
 *  <li>PhraseFilter3</li>
 *  <li>PhraseFilter2</li>
 *  <li>NGramFilter</li>
 * </ul>
 * @author asbjorn
 *
 */

public class DocumentAnalyzer extends Analyzer {
	private static Analyzer CONTENT_ANALYZER;
	public static final int INTER_ANCHOR_ANALYZER = 4;
	private static Analyzer ANCHOR_ANALYZER;
	private static boolean bigrams = false;
    private static File stopwordFile = new File("stop_words.txt");
	public static final HashSet<String> ENGLISH_STOP_WORDS = new HashSet<String>();
    
    public static enum Filters {NGramFilter, PhraseFilter2, PhraseFilter3};
    public Enum<Filters> filter;
	
	/**
     * Constructor which also loads the stop_words.<br/><br/>
     * <b>Note</b> that this constructore uses the PhraseFilter3 as default!<br/>
     * <b>Also note</b> that this constructore loads a file named 'stop_words.txt' 
     * in the current working directory. The 'stop_words.txt' file <b>must</b> have
     * one stopword for each line (# would imply a comment!).
     *
	 */
	public DocumentAnalyzer() {
		String stopwordFile = System.getProperty("willie.analyzer.stopwordFile");
		if (stopwordFile == null || stopwordFile.equals("")) init_stopWords(this.stopwordFile);
		else {
			System.out.println("Using system.property stopword file");
			init_stopWords(new File(stopwordFile)); 
		}
			
        CONTENT_ANALYZER = new SimpleAnalyzer();
        this.filter = Filters.PhraseFilter3;
	}
	
    /**
     * Provides the control of specifying which stop_word file to be used.<br/>
     * <b>Note</b> that it has to be a @see java.io.File object.
     * @param stopwordFile
     */
    public DocumentAnalyzer(File stopwordFile) {
        this();
        init_stopWords(stopwordFile);
    }
    
    /**
     * Provides the control of specifying wheter unigrams should be included in the 
     * index. Default is <b>true</b>.
     * @param unigrams
     */
	public DocumentAnalyzer(boolean unigrams) {
        this();
		if(unigrams) {
			this.bigrams = unigrams;
		}
	}
    
    /**
     * Provides the control of specifying which TokenFilter to use. Must 
     * choose between those defined in the @see DocumentAnalyzer#Filters
     * @param filters
     * @see org.apache.lucene.analysis.TokenFilter
     */
    public DocumentAnalyzer(Filters filters) {
        this();
        chooseFilter(filters);
    }
    
    /**
     * Provides the control of specifying both unigrams and stopword file.
     * @param unigrams
     * @param stopwordFile
     */
    public DocumentAnalyzer(boolean unigrams, File stopwordFile) {
        if(unigrams) {
            this.bigrams = unigrams;
        }
        init_stopWords(stopwordFile);
    }
    
    /**
     * Provides the control of specifying unigrams, stopword file and which TokenFilter 
     * to be used.
     * @param unigrams
     * @param stopwordFile
     * @param filters
     * @see org.apache.lucene.analysis.TokenFilter
     */
    public DocumentAnalyzer(boolean unigrams, File stopwordFile, Filters filters) {
        this(unigrams, stopwordFile);
        chooseFilter(filters);
        
    }
    
    public void setBigrams() {
    	this.bigrams = true;
    }
    
    private void init() {
        CONTENT_ANALYZER = new SimpleAnalyzer();
    }
    /**
     * This method chooses the TokenFilter based on the user-input.
     * @param f
     */
	private void chooseFilter(Filters f) {
        if(f == Filters.PhraseFilter3) {
            this.filter = Filters.PhraseFilter3;
        } else if(f == Filters.PhraseFilter2) {
            this.filter = Filters.PhraseFilter2;
        } else if(f == Filters.NGramFilter) {
            this.filter = Filters.NGramFilter;
        } else {
            this.filter = Filters.PhraseFilter3;
        }
    }
    
    /**
     * This method is called by the Constructor and loads in a list of stopwords 
     * from a file. The stopwords are stored as a HashSet, so we can maintain a 
     * very fast lookup. 
     *
     */
	private void init_stopWords(File stopwordFile) {
		try {
            if (stopwordFile == null) {
                stopwordFile = this.stopwordFile;
            }
			File tmpFil = stopwordFile;
			BufferedReader in = new BufferedReader(new FileReader(tmpFil));
			String tmpWord;
			while((tmpWord = in.readLine()) != null) {
				if(tmpWord.startsWith("#") || tmpWord.startsWith(" ")){
					continue;
				}
				ENGLISH_STOP_WORDS.add(tmpWord);
			}
			in.close();
		} catch(IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Returns the number of stopwords which has been loaded. If you want to see each stopword 
	 * then you can access thish HashSet directly.
	 * 
	 * @return int number of stopwords in memory
	 */
	public int numberOfStopwords() {
		return ENGLISH_STOP_WORDS.size();
	}
	
    /**
     * This method is necessary to provide the requested tokenStream.<br/>
     * It's basicly a wrapper for the underlying TokenFilter.
     */
	public TokenStream tokenStream(String fieldName, Reader reader) {
		//return analyzer.tokenStream(fieldName, reader)
		//return new PhraseFilter2(new SimpleAnalyzer().tokenStream(fieldName, reader),2);
		//(new SimpleAnalyzer().tokenStream(fieldName, reader),2);
        
        //return new PhraseFilter3(new SimpleAnalyzer().tokenStream(fieldName, reader),2);
        
        TokenStream ts = null;
        
        if (this.filter == Filters.PhraseFilter3) {
            ts = new PhraseFilter3(new SimpleAnalyzer().tokenStream(fieldName, reader),2);
        } else if(this.filter == Filters.PhraseFilter2) {
            ts = new PhraseFilter2(new SimpleAnalyzer().tokenStream(fieldName, reader),2);
        } else if(this.filter == Filters.NGramFilter) {
            ts = new NGramFilter(new SimpleAnalyzer().tokenStream(fieldName, reader),2);
        }
        return ts;
	}
	
	/*
	 * My own test phrasefilter which creates special bigrams and unigrams!
	 */
	private static class PhraseFilter3 extends TokenFilter {
		protected UnboundedFifoBuffer outputBuf;
		protected UnboundedFifoBuffer tokenBuf;
		protected CircularFifoBuffer ngramBuf;
		protected Token lastToken;
		protected Token nextToken;
		protected Token prevToken;
		protected Token mainToken;
		protected Token token;
		protected Token ngramToken = new Token();
		protected boolean gotStopword = false;
		protected String type = "ngrams";
		protected StringBuilder messString;
		protected boolean outputUnigrams = true;
		
		public PhraseFilter3(TokenStream input, int size) {
			super(input);
			outputBuf = new UnboundedFifoBuffer();
			tokenBuf = new UnboundedFifoBuffer();
			ngramBuf = new CircularFifoBuffer(size);
			
		}
		
		public boolean isStopWord(Token token) {
			if( ENGLISH_STOP_WORDS.contains( String.valueOf(token.termBuffer()).trim() ) ) {
				return true;
			}
			return false;
		}
		
		/*
		 * Fills the outputBuf with tokens (both unigrams and bigrams!).
		 */
		public void fillOutputBuf() throws IOException{
			boolean addedToken = false;
			do {
				token = this.getNextToken();
				if (token != null) {
					ngramBuf.add(token);
					//System.out.println("Adds token: "+token);
					lastToken = token;
					addedToken = true;
				} else {
					break;
				}
			} while(!ngramBuf.isFull());
			
			/*
			 * If no new token could be added to the ngram buffer, we have 
			 * reached the end of the input stream and have to discard the 
			 * least recent (ngram)token.
			 */
			if(!addedToken) {
				if(ngramBuf.isEmpty()) {
					return;
				} else {
					ngramBuf.remove();
				}
			}
			
			/* Ok, ngramBuf is full. Now, iterate over all tokens in ngramBuf and create unigrams */
			
			/* This is the part where we identify stopwords, and create bigrams of the 
			 * prevToken + stopword and the stopword (prevToken) + next token. */
			boolean addedBigram = false;
			int i=0;
			Token ngram;
			//messString = new StringBuilder();
			
			for(Iterator<Token> it = ngramBuf.iterator(); it.hasNext();) {
				token = it.next();
				//if(prevToken != null) System.out.println("prevToken: "+prevToken+". Is stopword? "+(isStopWord(prevToken)? "yes" : "no"));
				
				
				if(gotStopword && prevToken != null) { 
					/* prevToken was a stopword, adds prev+this token!!*/
					//testNGramBuf.add(prevToken.termText()+" "+token.termText());
					//System.out.println("Has a prevToken which is a stopword, and a new token!");
					ngram = new Token(prevToken.termText()+" "+token.termText(), 
							prevToken.startOffset(), // start offset 
							token.endOffset(), // end offset
							this.type);
					
					//System.out.println("added ngram2: "+ngram);

//					ngramToken.setTermLength(tmp.length);
//					ngramToken.setTermBuffer( tmp, 
//							prevToken.startOffset(), 
//							token.endOffset());
					//ngram.setPositionIncrement(0);
					//addBigram(ngram);
					
					//ngramToken.setPositionIncrement(0);
					outputBuf.add(ngram);
					
					//System.out.println("New Bigram: "+prevToken.termText()+" "+token.termText());
					gotStopword = false;
					addedBigram = true;
				}

				if (isStopWord(token)) {
					// we got a stopword! Start a bigram
					gotStopword = true;
				} else {
                    // add non-stopword unigrams to index!
					if(outputUnigrams) {
                        outputBuf.add(token); 
					}
				}

				if(gotStopword && prevToken != null) {
					// found a stopword in current token, and we have 
					// a previouse token! Create a bigram (prevToken+token)!
					
					//System.out.println("prevToken is != null! gotStopword: "+(gotStopword? "yes" : "no")+" token: "+token.termText()+" prevToken: "+prevToken.termText());
					
					if(!addedBigram) {
						
						ngram = new Token(prevToken.termText()+" "+token.termText(), 
								prevToken.startOffset(), // start offset 
								token.endOffset(), // end offset
								this.type);
						//System.out.println("added ngram: "+ngram+" gotStopword: "+gotStopword);
						
//						messString = new StringBuilder();
//						messString.append(prevToken.termBuffer());
//						messString.append(" ");
//						messString.append(token.termBuffer());
//						char[] tmp = messString.toString().toCharArray();
//						
//						ngramToken.setTermLength(tmp.length);
//						ngramToken.setTermBuffer( tmp, 
//								prevToken.startOffset(), 
//								token.endOffset());
//						System.out.println("New ngram: "+ngramToken);
						outputBuf.add(ngram);
						//ngram.setPositionIncrement(0);
					}
				}
				
				prevToken = token;
				i++;
				it.remove();
				addedBigram = false;
			}

		}
		
		public Token getNextToken() throws IOException {
			if(tokenBuf.isEmpty()) {
				nextToken = input.next();
				
				/* if lastToken's termText has length == 1, then fetch next! */
				/*
				while(lastToken.termText().length() == 1) {
					lastToken = input.next();
				} */
				
				if(nextToken != null)  {
					tokenBuf.add(nextToken);
					return this.getNextToken();
				} else {
					/* no more tokens from TokenStream input (end-of-stream) */
					return null;
				}
			} else {
				return (Token) tokenBuf.remove();
			}
		}
		
		public Token next() throws IOException {
			if(outputBuf.isEmpty()) {
				this.fillOutputBuf();
			}
			if (outputBuf.isEmpty()) {
				// If still empty?
				return null;
			} else {
				//Token t = (Token)outputBuf.remove();
				//System.out.println("Token: "+t);
				//return t;
				return (Token) outputBuf.remove();
			}
		}
	}
	
	
	
	
	
	private static class PhraseFilter2 extends TokenFilter {
		protected CircularFifoBuffer ngramBuf;
		protected UnboundedFifoBuffer outputBuf;
		protected UnboundedFifoBuffer tokenBuf;
		
		/* test buffer */
		protected UnboundedFifoBuffer testNGramBuf;
		
		protected StringBuffer[] ngrams;
		protected String type = "ngrams";
		protected Token prevToken = null;
		protected Token lastToken;
		public final static int DEFAULT_NGRAM_SIZE = 2;
		protected boolean outputUnigrams = true;
		protected boolean firstRun = true;
		
		public PhraseFilter2(TokenStream input, int ngramSize) {
			super(input);
			this.outputBuf = new UnboundedFifoBuffer();
			this.tokenBuf = new UnboundedFifoBuffer();
			this.testNGramBuf = new UnboundedFifoBuffer();
			if(ngramSize > 0) ngramSize++;
			this.setNGramSize(ngramSize);
		}
		
		public PhraseFilter2(TokenStream input) {
			this(input, DEFAULT_NGRAM_SIZE);
		}
		
		public PhraseFilter2(TokenStream input, String fieldName) {
			this(input, DEFAULT_NGRAM_SIZE);
			this.setType(fieldName);
		}
		
		
		public PhraseFilter2 setType(String type) {
			this.type = type;
			return this;
		}
		
		public PhraseFilter2 setOutputUnigrams(boolean flag) {
			this.outputUnigrams = flag;
			return this;
		}
		
		public PhraseFilter2 setNGramSize(int n) {
			if(n < 1) {
				throw new IllegalArgumentException("N-gram size must be >= 1");
			}
			this.ngramBuf = new CircularFifoBuffer(n);
			this.ngrams = new StringBuffer[n];
			
			for(int i=0; i<this.ngrams.length; i++) {
				this.ngrams[i] = new StringBuffer();
			}
			return this;
		}
		
		protected void clearNGrams() {
			for(int i=0; i<this.ngrams.length; i++) {
				this.ngrams[i].setLength(0);
			}
		}
		
		public Token next() throws IOException {
			if(outputBuf.isEmpty()) {
				this.fillOutputBuf();
			}
			if (outputBuf.isEmpty()) {
				// If still empty?
				return null;
			} else {
				//Token t = (Token)outputBuf.remove();
				//System.out.println("token: "+t.termText());
				//return t;
				return (Token) outputBuf.remove();
			}
		}
		
		
		/*
		 * This method identifies tokens (and ngram's).
		 */
		public Token getNextToken() throws IOException {
			if(tokenBuf.isEmpty()) {
				Token lastToken = input.next();
				if(lastToken != null)  {
					/*for(int i=1; i<lastToken.getPositionIncrement(); i++) {
						tokenBuf.add(new Token("_", lastToken.startOffset(), lastToken.startOffset()));
					}*/
					tokenBuf.add(lastToken);
					return this.getNextToken();
				} else {
					/* no more tokens from TokenStream input (end-of-stream) */
					return null;
				}
			} else {
				return (Token) tokenBuf.remove();
			}
		}
		
		public boolean addBigram(Token ngram) {
			Token tmp = null;
			for(Iterator<Token> i = testNGramBuf.iterator(); i.hasNext(); ) {
				tmp = (Token) i.next();
				
				// If the buffer already contains the bigram, don't add a new one!
				if((tmp.termText()).equals(ngram.termText())) {
					//System.out.println("\t'"+tmp.termText()+"' equals '"+ngram.termText()+"'");
					return false;
				}
			}
			//System.out.println("\tAdds: "+ngram.termText());
			testNGramBuf.add(ngram);
			return true;
		}
		
		/*
		 * It's this method that fills the ngramBuffer. The method that 
		 * identify ngram's is the @link{getNextToken}.
		 */
		public void fillOutputBuf() throws IOException {
			boolean addedToken = false;
			do {
				Token token = this.getNextToken();
				if (token != null) {
					//System.out.println("ngramBuf-append: "+token.termText());
					ngramBuf.add(token);
					lastToken = token;
					addedToken = true;
				} else {
					break;
				}
			} while(!ngramBuf.isFull());
			
			// prints out the ngramBuf
			System.out.print("ngramBuf: ");
			Token tmp = null;
			for(Iterator i = ngramBuf.iterator(); i.hasNext(); ) {
				tmp = (Token)i.next();
				System.out.print(tmp.termText()+" ");
			}
			System.out.println();
			

			/*
			 * If no new token could be added to the ngram buffer, we have 
			 * reached the end of the input stream and have to discard the 
			 * least recent (ngram)token.
			 */
			if(!addedToken) {
				if(ngramBuf.isEmpty()) {
					return;
				} else {
					ngramBuf.remove();
				}
			}
			this.clearNGrams();
			
			/* Creates a 2D-matrix with init values equal 0 */
			int[] endOffsets = new int[ngramBuf.size()];
			for(int i=0; i<endOffsets.length; i++) {
				endOffsets[i] = 0;
			}
			
			if(!firstRun) {
				// removes the first object in ngramBuf!
				//System.out.println("Will remove: "+ngramBuf.get());
				ngramBuf.remove();
			}
			
			/*
			 * This adds the tokens to the 'ngram' StringBuffer!
			 * f.ex: 'the','network' is appended in this way:
			 * 		ngram.append('the')
			 * 		ngram.append('the network')
			 */
			int i=0;
			Token token = null;
			boolean gotStopword = false;
			boolean addedBigram = false;
			for(Iterator it = ngramBuf.iterator(); it.hasNext();) {
				token = (Token) it.next();
				//System.out.println("LOLtoken: "+token.termText());
				
				if(gotStopword && prevToken != null) { 
					/* prevToken was a stopword, adds prev+this token!!*/
					//testNGramBuf.add(prevToken.termText()+" "+token.termText());
					
					Token ngram = new Token(prevToken.termText()+" "+token.termText(), 
							prevToken.startOffset(), // start offset 
							token.endOffset(), // end offset
							this.type);
					addBigram(ngram);
					
					//System.out.println("New Bigram: "+prevToken.termText()+" "+token.termText());
					gotStopword = false;
					addedBigram = true;
				}
				
				for(String stopword : ENGLISH_STOP_WORDS) {
					if ((token.termText()).equals(stopword)) {
						// we got a stopword! Start a bigram
						gotStopword = true;
					}
				}
				
				if(gotStopword && prevToken != null) {
					// found a stopword in current token, and we have 
					// a previouse token! Create a bigram!
					//testNGramBuf.add(token.termText()+" "+prevToken.termText());
					if(!addedBigram) {
						
						Token ngram = new Token(prevToken.termText()+" "+token.termText(), 
								prevToken.startOffset(), // start offset 
								token.endOffset(), // end offset
								this.type);
						addBigram(ngram);
						
						//System.out.println("NNNew Bigram: "+prevToken.termText()+" "+token.termText());
					}
					//System.out.println("New Bigram: "+prevToken.termText()+" "+token.termText());
				}
				
				prevToken = token;
				addedBigram = false;
				
				for(int j=i; j<ngrams.length; j++) {
					if(ngrams[j].length() != 0) {
						ngrams[j].append(" ");
					}
					ngrams[j].append(token.termText());
				}
				
				endOffsets[i] = token.endOffset();
				i++;
			}
			prevToken = null;
			firstRun = false;
			
			
			/* This one adds the unigram to the tokenStream (outputBuf!) */
			//System.out.println("ngramBuf size: "+ngramBuf.size());
			
			if( (!ngramBuf.isEmpty()) && outputUnigrams) {
				Token unigram = null;
				for(Iterator<Token> it = ngramBuf.iterator(); it.hasNext(); ) {
					unigram = it.next();
					gotStopword = false;
					
					sloop: 
						for(String stopword : ENGLISH_STOP_WORDS) {
						if(stopword.equals(unigram.termText())) {
							gotStopword = true;
							break sloop;
						} 
					}
					if (!gotStopword) {
						unigram.setPositionIncrement(1);
						System.out.println("Added Unigram: "+unigram.termText());
						//outputBuf.add(unigram);
						addBigram(unigram);
					} 
					
				}
				//System.out.println("Checking if '"+unigram.termText()+"' is a unigram candidate!");
			} 
			
			Token haha = null;
			for(Iterator<Token> t = testNGramBuf.iterator(); t.hasNext(); ) {
				haha = t.next();
				if (!outputUnigrams) {
					haha.setPositionIncrement(1);
				} else {
					haha.setPositionIncrement(0);
				}
				//System.out.println("Bigram: "+haha); //+" (start:"+haha.startOffset()+", stop:"+haha.endOffset()+")");
				outputBuf.add(haha);
				//t.remove();
			}
			
			testNGramBuf.clear();
		}
	}
	
	
	public class NGramFilter extends TokenFilter {

		protected CircularFifoBuffer ngramBuf;
		protected UnboundedFifoBuffer outputBuf;
		protected UnboundedFifoBuffer tokenBuf;
        protected StringBuffer[] ngrams;
		protected String type = "ngram";

		/**
		 * default n-gram size is 2.
		 */
		public final static int DEFAULTNGRAMSIZE = 2;

	  /**
		 * By default, we output unigrams as well as n-grams.
		 */
		protected boolean outputUnigrams = true;

	  /**
		 * Construct an NGramFilter with the specified n-gram size from the 
		 * TokenStream <code>input</code>
		 * 
		 * @param input input stream
		 * @param ngramsize maximum n-gram size produced by the filter.
		 */
		public NGramFilter(TokenStream input, int ngramsize) {
			super(input);
			this.outputBuf = new UnboundedFifoBuffer();
			this.tokenBuf = new UnboundedFifoBuffer();
			this.setNGramSize(ngramsize);
		}
		
	  /**
	   * Construct an NGramFilter with default n-gram size.
	   *  
	   * @param input input stream
	   */
		public NGramFilter(TokenStream input) {
			this(input, DEFAULTNGRAMSIZE);
		}

	  /**
	   * Construct an NGramFilter with the specified field name for n-gram tokens. 
	   * 
	   * @param input input stream
	   * @param fieldname field name for n-gram tokens
	   */
		public NGramFilter(TokenStream input, String fieldname) {
			this(input, DEFAULTNGRAMSIZE);
			this.setType(fieldname);
		}
		
		/**
		 * Set the type of the n-gram tokens produced by this filter.
		 * (default: "ngram")
		 * 
		 * @param type token type
		 * @return receiver
		 */
		public NGramFilter setType(String type) {
			this.type = type;
			return this;
		}
		
		/**
		 * Shall the output stream contain the input tokens (unigrams) as well as
		 * n-grams? (default: true.)
		 * 
		 * @param flag
		 * @return receiver
		 */
		public NGramFilter setOutputUnigrams(boolean flag) {
			this.outputUnigrams = flag;
			return this;
		}
		
		/**
		 * Set the n-gram size (default: 2)
		 * 
		 * @param n n-gram size
		 * @return receiver
		 */
		public NGramFilter setNGramSize(int n) {
			if (n < 1) {
				throw new IllegalArgumentException("N-gram size must be >= 1");
			}
			this.ngramBuf = new CircularFifoBuffer(n);
	    this.ngrams = new StringBuffer[n];
	    for (int i = 0; i < this.ngrams.length; i++) {
	      this.ngrams[i] = new StringBuffer();
	    }
			return this;
		}
	  
	  /**
	   * Clear the StringBuffers that are used for storing the output n-grams.
	   */
	  protected void clearNGrams() {
	    for (int i = 0; i < this.ngrams.length; i++) {
	      this.ngrams[i].setLength(0);
	    }
	  }
		
		/* (non-Javadoc)
		 * @see org.apache.lucene.analysis.TokenStream#next()
		 */
		public Token next() throws IOException {
			if (outputBuf.isEmpty()) {
				this.fillOutputBuf();
			}
			if (outputBuf // is still empty
					.isEmpty()) {
				return null;
			} else {
				return (Token) outputBuf.remove();
			}
		}

		/**
		 * Get the next token from the input stream and push it on the token buffer.
	   * If we encounter a token with position increment > 1, we put filler tokens 
	   * on the token buffer.
	   * 
	   * Returns null when the end of the input stream is reached. 
		 */
		private Token getNextToken() throws IOException {
			if (tokenBuf.isEmpty()) {
				Token lastToken = input.next();
				if (lastToken != null) {
					for (int i = 1; i < lastToken.getPositionIncrement(); i++) {
						tokenBuf.add(new Token("_", lastToken.startOffset(), lastToken.startOffset()));
					}
					tokenBuf.add(lastToken);
					return this.getNextToken();
				} else {
					return null;
				}
			} else {
				return (Token) tokenBuf.remove();
			}
		}
		
	  /**
	   * Fill the output buffer with new n-grams.
	   * 
	   * @throws IOException
	   */
		private void fillOutputBuf() throws IOException {
			boolean addedToken = false;
			
			/*
	     * Try to fill the ngram buffer.
			 */
	    do {
				Token token = this.getNextToken();
				if (token != null) {
					ngramBuf.add(token);
					addedToken = true;
				} else {
					break;
				}
			} while (! ngramBuf.isFull());
			
	    /*
	     * If no new token could be added to the ngram buffer, we have reached
	     * the end of the input stream and have to discard the least recent token.
	     */
			if (! addedToken) {
				if (ngramBuf.isEmpty()) {
					return; 
				} else {
					ngramBuf.remove();
				}
			}

	    this.clearNGrams();
	 
	    int[] endOffsets = new int[ngramBuf.size()];
			for (int i = 0; i < endOffsets.length; i++) {
				endOffsets[i] = 0;
			}
			
	    int i = 0;
			Token token = null;
			for (Iterator it = ngramBuf.iterator(); it.hasNext(); ) {
				token = (Token) it.next();
				for (int j = i; j < ngrams.length; j++) {
					if (ngrams[j].length() != 0) {
						ngrams[j].append(" ");
					}
					ngrams[j].append(token.termText());
				}
				endOffsets[i] = token.endOffset();
				i++;
			}
			
			if ((! ngramBuf.isEmpty()) && outputUnigrams) {
				Token unigram = (Token) ngramBuf.get();
				unigram.setPositionIncrement(1);
				outputBuf.add(unigram);
			}
			
			/*
	     * Push new tokens to the output buffer.
			 */
			for (int j = 1; j < ngramBuf.size(); j++) {
				Token ngram = new Token(ngrams[j].toString(), 
						((Token) ngramBuf.get()).startOffset(), 
						endOffsets[j],
						this.type);
				if ((! outputUnigrams) && j == 1) {
					ngram.setPositionIncrement(1);
				} else {
					ngram.setPositionIncrement(0);
				}
				outputBuf.add(ngram);
			}
		}

	}
	
	/* For debugging purpose, only! */
	public static void main(String[] args) throws Exception {
		StringBuffer text = new StringBuffer();
		
		if (args.length == 1) {
			System.out.println("fil: "+args[0]);
			System.exit(0);
		} else {
			for(int i=0; i<args.length; i++) {
				text.append(args[i]);
				text.append(' ');
			}
			System.out.println("Input: "+text);
			System.out.println();
		}
		
		DocumentAnalyzer test = new DocumentAnalyzer(DocumentAnalyzer.Filters.PhraseFilter3);
		test.setBigrams();
		
		
		TokenStream ts = test.tokenStream("debug", new StringReader(text.toString()));
		
		Token token;
		while((token = ts.next()) != null) {
			System.out.println("Token: "+token+" posInc:"+token.getPositionIncrement());
		}
		System.out.println("Ferdig...");
	}
}
