package baldr;

import it.unimi.dsi.mg4j.document.CompositeDocumentFactory;
import it.unimi.dsi.mg4j.document.Document;
import it.unimi.dsi.mg4j.document.DocumentFactory;
import it.unimi.dsi.mg4j.document.DocumentIterator;
import it.unimi.dsi.mg4j.document.HtmlDocumentFactory;
import it.unimi.dsi.mg4j.document.PropertyBasedDocumentFactory;
import it.unimi.dsi.mg4j.document.TRECDocumentCollection;
import it.unimi.dsi.mg4j.document.TRECHeaderDocumentFactory;
import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.Properties;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.collections.Buffer;
import org.apache.commons.collections.buffer.BoundedFifoBuffer;
import org.apache.commons.collections.buffer.CircularFifoBuffer;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import baldr.interfaces.IndexImpl;


/**
 * 
 * @author fellingh
 * This is the Gov2Parser which, as the name says, parses GOV2 documents and outputs their 
 * content in a structural manner. 
 *  <b>This parser is initialized as a Thread, since it implements Runnable</b>.
 *  
 *  The sequence of actions is as follows:
 *  <ul>
 *  	<li>Fetches GOV2 file from filesytem</li>
 *  	<li>MG4J is passed an array of files to parse</li>
 *  	<li>MG4J returns an iterator</li>
 *  	<li>Gov2Parser uses the iterator to extract data</li>
 *  	<li>When internal buffer is full, then pass buffer to Indexer.</li>
 *  	<li>When internal buffer is unboundled by the Indexer, we'll continue to iterate until the iterator returns empty</li>
 *  	<li>If no more docs is extracted, then give MG4J a new array of files to parse, and so on....</li>
 *  </ul>
 *  
 *  @see	Indexer
 */

public class Gov2Parser implements Runnable {
    /*
     * Tar inn en "path" som input, og tar s  henter ut alle filer og 
     * parser de som TREC filer.
     */
    File dir;
    File[] filer;
    ArrayList<Gov2Document> documents = new ArrayList<Gov2Document>();
    List<String> files = new ArrayList<String>();
    
    /* Buffers */
    BoundedFifoBuffer gov2Docs;
    BoundedFifoBuffer readyGov2Docs, freeGov2Docs;
    
    //Buffer gov2Buffer;
    //= new CircularFifoBuffer(5000); /**/
    //CircularFifoBuffer tmpDocs = new CircularFifoBuffer(1000);
    //BoundedFifoBuffer tmpDocs = new BoundedFifoBuffer(1000);

    public boolean parsing = true;
    public boolean collect = false;
    private boolean debug = true;
    private Gov2Document newDoc;
    private Gov2Document2 doc;
    private int maxNumFiles = 10;
    private int numFiles = 0;
    private int numDocs = 0;
    private int maxNumDocs = 1100; /* antall <doc>'s vi har i arrayet! */
    private String dirName;
    private IndexImpl indexer;
    private Properties p = new Properties();
    private TRECHeaderDocumentFactory thd = new TRECHeaderDocumentFactory();
    private HtmlDocumentFactory hdf;
    private DocumentFactory docFac;
    private TRECDocumentCollection d;
    private Logger logger = Logger.getLogger(this.getClass());
    
    /* Die after reaching this upper file limit! */
    //private int dieAfter = 70;

    
    public Gov2Parser(List<String> files, IndexImpl in, final BoundedFifoBuffer gov2Docs) {
        //this.dirName = dirName;
    	this.files = files;
        this.indexer = in;
        this.gov2Docs = gov2Docs;
        //this.gov2Buffer = gov2Docs;
        logger.setLevel(Level.DEBUG);
        
        // Initializes the indexing thread, and feeds the gov2Docs doc queue!
        this.indexer.indexDocs();
        
        try {
        	p.addProperty(PropertyBasedDocumentFactory.MetadataKeys.ENCODING, new String("ISO-8859-1"));
        	hdf = new HtmlDocumentFactory( p );
        	DocumentFactory[] facArray = {thd, hdf};
        	docFac = CompositeDocumentFactory.getFactory(facArray, new String[] {"text","title","anchor"});
        }catch(Exception e) {
        	e.printStackTrace();
        }
    }
    
    /**
     * Sets the two buffers used to handle finished and ready Gov2Docs. This is an attempt to save us from the GC cost 
     * of actually create each Gov2Document. Now, we simply "warmup" by create 5000 docs.<br/>
     * 
     * </i>Think of it as an ObjectPool.</i>
     * @param readyGov2Docs
     * @param freeGov2Docs
     */
    public void setBuffers(final BoundedFifoBuffer readyGov2Docs, final BoundedFifoBuffer freeGov2Docs) {
    	this.readyGov2Docs = readyGov2Docs;
    	this.freeGov2Docs = freeGov2Docs;
    }
    
    @Deprecated
    private void fetchFiles(String dirName) {
    	File[] filer;
    	
    	File dir = new File(dirName);
    	filer = dir.listFiles();
    	for(File f : filer) {
    		/********* DEBUGGING **********/
    		if(files.size() == 75) break; /* Debugging!!!!*/
    		/********* DEBUGGING **********/
    		
    		if(f.isDirectory()) {
    			/* Recursion! Needed to iterate down the path hierarchy! */
    			fetchFiles(f.toString());
    			continue;
    		}
    		files.add(f.getAbsolutePath());
    	}
    }
    
    /**
     * This method starts the whole Thread. It basicly iterates over a predefined number of files, 
     * and then sends off a method request to this.indexFile with the sublist.<br/>
     * This loops until there are no more files to parse.
     */
    public void run() {
    	int last=0;
    	int numberOfFiles = files.size();
    	int iterations = 0;
    	long percentageComplete = 0;
    	
    	for(int i=0;i<files.size();i++) {
    		if (numFiles == maxNumFiles) {
    			indexFile( files.subList(last, i));
    			last = i;
    			numFiles = 0;

    			percentageComplete = (iterations / numberOfFiles)*100; // Gets a rough percentage estimation
    			logger.debug(Thread.currentThread().getName()+": Has completed "+percentageComplete+"% of its GOV2 files.");
    		}
    		iterations++;
    		numFiles++;
    	}
    	
    	indexFile( files.subList(last, files.size()));
        this.parsing = false;
        percentageComplete = (iterations / numberOfFiles)*100; // Gets a rough percentage estimation
		logger.debug(Thread.currentThread().getName()+": Has completed "+percentageComplete+"% of its GOV2 files.");
        logger.debug("Gov2Parser finished running..");
        
        // Very important! Tells the AddDocs thread while loop to stop!
        this.indexer.setDoneparsing(true);
    }

    
    /**
     * This is the method which actually parses the GOV2 files, and from there on passes it 
     * to the Indexer.
     * @param gov2File
     * @see Indexer#doIt2(CircularFifoBuffer)
     */
    private void indexFile(String[] gov2File) {
    	
    	try {
    		// Buffer size:
    		// 8192 bytes
    		// 16384 bytes
    		// 32768
    		// 65536
        	d = new TRECDocumentCollection(gov2File, docFac, 32768, false);
        	Document dt;        	
        	
        	DocumentIterator it = d.iterator();
        	MutableString title = new MutableString();
        	MutableString content = new MutableString();
        	FastBufferedReader r;
        	MutableString parsedContent;
        	int c;
        	char l, prev;
        	
        	do {
        		parsedContent = new MutableString();
        		dt = it.nextDocument();
        		
        		// Very important to break out if there is no more docs!
        		if (dt==null) break;
        		
        		r = (FastBufferedReader) dt.content(0);
        		r.readLine(content);
        		prev = ' ';
        		while( (c = r.read()) != -1) {
        			l = (char) c;
        			if (prev == ' ' && l == ' ') continue; // vil ikke ha repeterende mellomrom!
        			if (l == '\n' || l == '\r' || l=='\t') {
        				continue;
        			}
        		
        			parsedContent.append(l);
        			prev = l;
        		}
        		
        		r = (FastBufferedReader) dt.content(1);
        		r.readLine(title);
        		
        		// fetch free object from "free" buffer
        		synchronized (freeGov2Docs) {
        			while(freeGov2Docs.isEmpty()) {
        				freeGov2Docs.wait();
        			}
        			doc = (Gov2Document2) freeGov2Docs.remove();
        			doc.setDocno(dt.title().toString());
        			doc.setUri(dt.uri().toString());
        			doc.setHeadline(title.toString());
        			doc.setText(parsedContent.toString());
        			
        			//if(doc.document().getField("docno").stringValue().equals("GX004-83-0039097")) {
        			//	System.out.println("uri: "+doc.document().getField("uri").stringValue());
        			//	System.out.println("headline: "+doc.document().getField("headline").stringValue());
        			//	System.out.println("text: "+doc.document().getField("text").stringValue());
        			//}
        			
				}
        		
        		// adds the doc object into the "ready" buffer, ready to be indexed
        		synchronized (readyGov2Docs) {
					while(readyGov2Docs.isFull()) {
						readyGov2Docs.wait();
					}
					readyGov2Docs.add(doc);
				}
        		
        		/*
        		newDoc = new Gov2Document(dt.title().toString(), title.toString(), dt.uri().toString(), parsedContent.toString());
        		
        		synchronized (gov2Docs) {
					while(gov2Docs.isFull()) {
						wait();
					}
					gov2Docs.add(newDoc);
				} */
        		
        		/*
        		if(!tmpDocs.isFull()) {
        			tmpDocs.add(newDoc);
        		} else {
        			// Need to truncate tmpDocs!
        			synchronized (gov2Docs) {
        				// Need to verify that there is space!
						while( !((gov2Docs.maxSize() - gov2Docs.size()) > 1000)) {
							logger.debug("Waiting for the indexer to consume items in buffer! Current size: "+gov2Docs.size());
							gov2Docs.wait();
						}
						logger.debug(Thread.currentThread().getName()+": Appending tmpDocs ("+tmpDocs.size()+") to gov2Docs!");
						gov2Docs.addAll(tmpDocs);
						tmpDocs.clear();
					}
        		}*/
        		
        		/*
        		synchronized (gov2Docs) {
            		//this.documents.add(newDoc);
        			while (gov2Docs.isFull()) {
        				try {
        					if(this.debug) {
        						logger.debug("Waiting for Indexer to consume docs.. ("+gov2Docs.size()+"). Calling wait();");
        					}
        					gov2Docs.wait();
        				} catch(Exception e) {
        					e.printStackTrace();
        				}
        			}
        			this.gov2Docs.add(newDoc);
				} */
        		
        		
        	} while(true); // continues until the iterator returns null!
        	//logger.debug("We're done parsing the docs! Exiting thread!");
        } catch(Exception e) {
        	e.printStackTrace();
        }
    }
    
    /**
     * A wrapper method which accepts a single String to hold the GOV2 file path.
     * @param gov2File
     * @see Gov2Parser#indexFile(List)
     */
    private void indexFile(String gov2File) {
    	/*
    	 * Adapts the indexFile method where we pass in a String[] array. 
    	 * Here we use a single String.
    	 */
    	indexFile(new String[] {gov2File});
    }
    
    /**
     * A wrapper method which accepts a List<String> to hold the GOV2 files to parse.
     * @param files
     * @see Gov2Parser#indexFile(List)
     */
    private void indexFile(List<String> files) {
    	MutableString tmp = new MutableString();
    	String filesString;
    	String[] filesArray;
    	
    	for(String fil : files) {
    		tmp.append(fil+" ");
    	}
    	filesString = tmp.toString();
    	filesArray = filesString.split(" ");
    	
    	indexFile(filesArray);
    }
    
    /**
     * This method was previously ment to return the parsed documents.
     * <p>
     * <b>Since we now use two threads and is based on the communication between them, 
     * we no longer uses this method in the Gov2 parsing.</b>
     * @return {@link ArrayList}
     * @deprecated
     */
    public ArrayList<Gov2Document> getDocuments() {
        return this.documents;
    }
}
