package baldr;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.commons.collections.Buffer;
import org.apache.commons.collections.BufferUtils;
import org.apache.commons.collections.buffer.BoundedFifoBuffer;
import org.apache.commons.collections.buffer.CircularFifoBuffer;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.SimpleAnalyzer;
import org.apache.lucene.index.*;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;

import baldr.interfaces.IndexImpl;


/**
 * Module which does the primary indexing stuff. 
 * @author asbjorn
 *
 */
public class Indexer implements IndexImpl {
    private static Analyzer testAnalyzer = new SimpleAnalyzer();
    private static Analyzer analyzer;
    
	private static Directory directory;
	private static IndexWriter iwriter;
    
    /* katalog der Lucene index'en skal være */
	private static String indexDir = Definitions.indexDir_unigrams;
    
    /* katalog der dokumentene som skal indekseres befinner seg. */
	private static String baseDir = Definitions.baseDir;
	
	/*  Buffers */
	private BoundedFifoBuffer gov2Docs = new BoundedFifoBuffer(5000);
	
	private BoundedFifoBuffer readyGov2Docs = new BoundedFifoBuffer(5000);
	private BoundedFifoBuffer freeGov2Docs = new BoundedFifoBuffer(5000);
	
    private long timeStart, timeStop;
    private Thread parserThread1;
    private Thread parserThread2;
    private Runnable addDocs;
    private Thread addDocsThread;
    private Thread indexDocThread;
    private AddDocs addDoc;
    private Logger logger = Logger.getLogger(this.getClass());
    private static final boolean debug = true;
    private ArrayList<String> files = new ArrayList<String>();
    public boolean doneParsing = false;
    public boolean startIndexing = false;
	
    /**
     * This is the constructor for the Indexer class. This initializes the necesessary variables. Among 
     * others the {@link Indexer#directory} is set to represent the {@link Definitions#indexDir}.
     * <p>
     * <b>Also the critical {@see IndexWriter} is created, with the wanted analyzer and such.<b>
     * 
     * @author Asbjrn Alexander Fellinghaug
     *
     */
	public Indexer() {
        if(this.debug) {
            logger.setLevel(Level.DEBUG);
        } else {
            logger.setLevel(Level.WARN);
        }

	}
	
	/**
	 * Method to start the indexing process, and to initialize the Gov2Parser thread.
	 * @see Gov2Parser
	 */
	public void startIndex(String indexDir, Analyzer analyzer) {
		this.analyzer = analyzer;
		try {
			directory = FSDirectory.getDirectory(indexDir);
		} catch(IOException e) {
			e.printStackTrace();
			logger.error("Fikk ikke åpnet index-dir..Nå dør jeg..");
			System.exit(0);
		}
		try {
			iwriter = new IndexWriter(directory, analyzer, true);
		} catch(IOException e) {
			e.printStackTrace();
			logger.error("IOException ved indexwriter");
		}
		logger.debug("Has successfully set indexdir->"+indexDir+" and analyzer.");
		
		if(this.analyzer instanceof DocumentAnalyzer) {
			logger.debug("Using DocumentAnalyzer. # stopwords: "+((DocumentAnalyzer)analyzer).numberOfStopwords());
		}
		
		iwriter.setMaxFieldLength(180000);
		iwriter.setMaxBufferedDocs(10000);
		iwriter.setMergeFactor(100);
		iwriter.setRAMBufferSizeMB(150); // Flush'er på 100MB
		//iwriter.setUseCompoundFile(false); // having this true, takes 7-33% more resources/time
		
		timeStart = System.currentTimeMillis();
		
		// Fetch all the files in here first!
		fetchFiles(Definitions.trecFile);
		int size = files.size();
		
		if(size == 0) {
			logger.warn("There is zero files to parse. No need to start indexing!");
			return;
		}
		
		// Fills the "free" buffer
		fillBuffer();
		
		//List<String> liste1 = files.subList(0, (files.size()/2));
		//List<String> liste2 =  files.subList((files.size()/2)+1, files.size());
		
		// First partition the file list 50-50, then pass it to each object!
		Gov2Parser g2p1 = new Gov2Parser( files, this, gov2Docs);
		//Gov2Parser g2p2 = new Gov2Parser( liste2, this, gov2Docs);
		
		
		// Passes on bufferpools
		g2p1.setBuffers(readyGov2Docs, freeGov2Docs);
		
		parserThread1 = new Thread(g2p1);
		//parserThread2 = new Thread(g2p2);
		
		parserThread1.setName("Parser 1");
		//parserThread2.setName("Parser 2");
		
		parserThread1.start();
		//parserThread2.start();
		
		/*
		gov2Parser = new Gov2Parser(baseDir, this);
		traad = new Thread(gov2Parser);
		traad.start(); */
	}
	
	
	private void fillBuffer() {
		int bufferSize = freeGov2Docs.maxSize();
		for(int i=0; i<bufferSize;i++) {
			freeGov2Docs.add(new Gov2Document2());
		}
		logger.debug("Has created "+bufferSize+" documents in free buffer.");
	}
	
	
	private void fetchFiles(File trecFile) {	
		if(!trecFile.isFile()) {
			logger.error("Could not open: "+trecFile.getPath()+". Need this file to determine ");
			return;
		}
		File tmpFile;
		
		try {
			BufferedReader in = new BufferedReader(new FileReader(trecFile));
			String t = "";
			while( (t = in.readLine()) != null) {
				if (t.startsWith("#")) continue;
				files.add(Definitions.baseDir+File.separator+t.replace("\n", ""));
			}
		} catch(IOException e) {
			e.printStackTrace();
		}
		
	}
	
	private void fetchFiles(String dirName) {
    	File[] filer;
    	
    	File dir = new File(dirName);
    	filer = dir.listFiles();
    	for(File f : filer) {
    		if(files.size() == 1000) break; // Will only parse 1000 trec files (ca 17GB).
    		
    		if(f.isDirectory()) {
    			/* Recursion! Needed to iterate down the path hierarchy! */
    			fetchFiles(f.toString());
    			continue;
    		}
    		files.add(f.getAbsolutePath());
    	}
    }
	
	@Deprecated
	public void hasDocs(ArrayList<Trec96Document> docs) {
		if (docs == null) {
            logger.warn("Got null docs from TrecParser!");
			return ;
		}
		for(int i=0; i<docs.size(); i++) {
			try {
				iwriter.addDocument(docs.get(i).document());
				//System.out.println("Got "+iwriter.numRamDocs()+" in RAM, cached "+(iwriter.ramSizeInBytes()/1024)+" kb of index");
			} catch(IOException e) {
				logger.error("Corrupt index exception (or something else) occured..");
				e.printStackTrace();
			}
        }
	}
	
	public synchronized void setDoneparsing(boolean value) {
		this.doneParsing = value;
	}
	
	public void indexDocs() {
		if(indexDocThread == null) {
			addDoc = new AddDocs(this.gov2Docs, this.readyGov2Docs, this.freeGov2Docs);
			indexDocThread = new Thread(addDoc);
			logger.debug("Will startup docIndexer Thread!");
			indexDocThread.start();
			logger.debug("Is now "+Thread.activeCount()+" threads up and running!");
		}
	}
	
	private class AddDocs implements Runnable {
		private final BoundedFifoBuffer docQueue;
		private final BoundedFifoBuffer readyGov2Docs;
		private final BoundedFifoBuffer freeGov2Docs;
		private int x = 0;
		private Gov2Document2 doc;
		//private final Buffer docBuffer;
		
		public AddDocs(final BoundedFifoBuffer docs, 
				final BoundedFifoBuffer readyGov2Docs, 
				final BoundedFifoBuffer freeGov2Docs) {
			docQueue = docs;
			this.readyGov2Docs = readyGov2Docs;
			this.freeGov2Docs = freeGov2Docs;
		}
		
		private void indexDocuments() {
			synchronized (readyGov2Docs) {
				while(!readyGov2Docs.isEmpty()) {
					doc = (Gov2Document2) readyGov2Docs.remove();
					try {
						iwriter.addDocument(doc.document());
						x++;
					} catch(IOException e) {
						e.printStackTrace();
					}
					
					// puts the used objects in the "free" buffer, ready to be used again!
					synchronized (freeGov2Docs) {
						freeGov2Docs.add(doc);
					}
					
					if(readyGov2Docs.size() < 2000 && !doneParsing) {
						readyGov2Docs.notify();
					}
				}
				logger.debug("Notifying the Gov2Parser thread. Has added "+x+" docs");
			}
			
			/*
			synchronized (docQueue) {
				while(!docQueue.isEmpty()) {
					doc = (Gov2Document) docQueue.remove();
					try {
						iwriter.addDocument(doc.document());
					} catch(IOException e) {
						logger.error("Corrupt index exception (or something else) occured.");
						e.printStackTrace();
					}
					x++;
					
					if(docQueue.size() < 2000 && !doneParsing) {
						//logger.debug("Notifying the Gov2Parser thread to continue with the GOV2 parsing..");
						docQueue.notify();
					}
				}
				logger.debug("Have added "+x+" documents to index!");
				x=0;
			} */
		}
			
		public void run() {
			while(!doneParsing || parserThread1.isAlive() || readyGov2Docs.isEmpty()) {
				if(readyGov2Docs.isEmpty()) {
					try {
						Thread.sleep(2000);
					} catch(InterruptedException e) {
						e.printStackTrace();
					}
					continue;
				}
				
				indexDocuments();
			}
			while(readyGov2Docs.size() > 0) {
				indexDocuments();
			}
			logger.debug("Run loop finish. ready size: "+readyGov2Docs.size()+" free size: "+freeGov2Docs.size());
			/*
			while(!doneParsing || parserThread1.isAlive() || docQueue.isEmpty() ) {
				// If docQueue is empty, then wait and check again!
				if (docQueue.isEmpty()) {
					try {
						//logger.debug("docQueue is empty. Sleeping for a while..");
						Thread.sleep(2000);
					} catch(InterruptedException e) {
						e.printStackTrace();
					}
					continue;
				}
				if(doneParsing) {
					logger.debug("parserThread1 isAlive: "+parserThread1.isAlive());
					//logger.debug("parserThread2 isAlive: "+parserThread2.isAlive());
				}
				//logger.debug("Indexing "+docQueue.size()+" documents!");
				indexDocuments();
			}
			
			// for some reason, I need this to consume "leftover" buffer items..
			while(docQueue.size() > 0) {
				indexDocuments();
			} */
			
			// Will exit this thread..
			//logger.debug("Run loop finish. DocQueue is: "+docQueue.size()+". parserThread1 is "+ 
			//		(parserThread1.isAlive()? "alive":"dead")+". Will now call doneParsing().");
			
			doneParsing();
		}
	}
	
	/**
	 * Calculates the size of all the TREC files readed.
	 * @return
	 */
	private long sizeOfGov2FilesReaded() {
		long sum=0;
		File gov2File;
		for(String file : files) {
			gov2File = new File(file);
			sum = sum + gov2File.length();
		}
		return sum;
	}
	
	/**
	 * Method that is to be called when the whole indexing process is completed. Here all the 
	 * necessary work to safely quit is done.
	 * @see Indexer#close()
	 */
	public void doneParsing() {
        timeStop = System.currentTimeMillis();
        long timeItTook = (timeStop - timeStart)/1000;
        long bytesReaded = sizeOfGov2FilesReaded();

        int seconds=0;
		int minutes=0;
		
		if (timeItTook > 60) {
			minutes = ((int)timeItTook) / 60;
			seconds = ((int)timeItTook)-(minutes*60);
		} else {
			seconds = (int)timeItTook;
		}
		

		
        System.out.println("Det tok "+minutes+" mins and "+seconds+" secs å indeksere.");
        System.out.println("Det er nå: "+iwriter.docCount()+" i index.");
        System.out.println("timeItTook: "+(timeStop-timeStart));
        System.out.println("Ferdig med  indeksere!\n===========================\n");
        System.out.println("We read "+bytesReaded+" bytes of GOV2 files.");
        close();
        System.out.println("Avslutter..");
	}
	
	/**
	 * Method to close the indexWriter. Throws an IOException if there is any error during 
	 * the process.  
	 * @throws IOException
	 */
	private void close() {
		try {
			iwriter.close();
		} catch(IOException e) {
			e.printStackTrace();
			logger.error("Klarte ikke å lukke indexwriter.. Mulig den allerede er lukket!");
		}
	}
	
}
